#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <time.h>

#include "file_matrix.h"

#define PRINT_HELP printf("Usage:\n"\
						  "\tGenerate a matrix (rows*cols) into 'filename'\n"\
						  "\t\t%s -g <filename> -r <rows> -c <cols>\n"\
						  "\tPrint the matrix from 'filename'\n"\
						  "\t\t%s -p <filename>\n"\
						  "\tCheck if the matrix from 'filename' is correct\n"\
						  "\t\t%s -t <filename>\n\n", argv[0], argv[0], argv[0]);

/* Individua il valore dell'elemento [i][j] della matrice */
#define M(i,j) (i+j)

int matrix_generate(const char *filename, int r, int c) {
  int fd;
  struct matrix_header mh;

  /* Crea il file per salvare la matrice */
  if ((fd = creat(filename, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) < 0) {
	perror(filename);
	return -1;
  }

  mh.rows = r;
  mh.cols = c;

  /* Salva l'header sul file */
  if (matrix_save(fd, &mh)) {
	close(fd);
	return -1;
  }

  /* Salva gli elementi sul file */
  for (r = 0; r < mh.rows; r++) {
	for (c = 0; c < mh.cols; c++) {
	  if (matrix_put_next_element(fd, M(r,c))) {
		close(fd);
		return -1;
	  }
	}
  }

  /* Chiude il file */
  if (close(fd)) {
	perror(filename);
  }
  
  return 0;
}

#define TESTS 50000

/* Ritorna un numero (random) tra 0 e (max-1) */
#define RAND(max) ((int) ((max) * (rand() / (RAND_MAX + 1.0))))

int matrix_check(const char *filename) {
  int fd, value, i, j, tests;
  struct matrix_header mh;

  /* Apre il file per leggere la matrice */
  if ((fd = open(filename, O_RDONLY)) < 0) {
	perror(filename);
	return -1;
  }

  /* Carica l'header dal file */
  if (matrix_load(fd, &mh)) {
	close(fd);
	return -1;
  }

  /* Imposta il seme */
  srand((unsigned int) time(NULL));

  /* Esegue il test randomico */
  tests = TESTS;
  while (tests--) {
	i = RAND(mh.rows);
	j = RAND(mh.cols);
	if (matrix_get_element(fd, &mh, i, j, &value)) {
	  close(fd);
	  return -1;
	}
	if (value != M(i,j)) {
	  printf("Error: [%03d][%03d] is %d instead of %d\n", i, j, value, M(i,j));
	  return -1;
	}
  }

  printf("All tests (%d) are passed successfully\n", TESTS);

  /* Chiude il file */
  if (close(fd)) {
	perror(filename);
  }

  return 0;
}

int matrix_print(const char *filename) {
  int fd, value, i, j;
  struct matrix_header mh;

  /* Apre il file per leggere la matrice */
  if ((fd = open(filename, O_RDONLY)) < 0) {
	perror(filename);
	return -1;
  }

  /* Carica l'header dal file */
  if (matrix_load(fd, &mh)) {
	close(fd);
	return -1;
  }

  /* Stampa gli elementi */
  for (i = 0; i < mh.rows; i++) {
	for (j = 0; j < mh.cols; j++) {
	  if (matrix_get_next_element(fd, &value)) {
		close(fd);
		return -1;
	  }
	  printf("[%03d][%03d]: %d\n", i, j, value);
	}
  }

  /* Chiude il file */
  if (close(fd)) {
	perror(filename);
  }

  return 0;
}

int main(int argc, char **argv) {
  int c;
  char op = (char)0; /* Tipo di operazione */
  int rows = 0, cols = 0;
  char *file;

  while ((c = getopt(argc, argv, "t:p:g:r:c:h")) != -1) {
	switch (c) {
	case 't':
	  if (op) {
		PRINT_HELP;
		return -1;
	  }
	  op = 't';
	  file = optarg;
	  break;

	case 'p':
	  if (op) {
		PRINT_HELP;
		return -1;
	  }
	  op = 'p';
	  file = optarg;
	  break;

	case 'g':
	  if (op) {
		PRINT_HELP;
		return -1;
	  }
	  op = 'g';
	  file = optarg;
	  break;

	case 'r':
	  rows = atoi(optarg);
	  break;

	case 'c':
	  cols = atoi(optarg);
	  break;
	  
	case 'h':
	  PRINT_HELP;
	  return 0;

	default:
	  printf("getopt returned character code 0%o\n", c);
	  return -1;
	}
  }

  switch (op) {
  case 't':
	return matrix_check(file);
  case 'p':
	return matrix_print(file);
  case 'g':
	if (cols > 0 && rows > 0) {
	  return matrix_generate(file, rows, cols);
	}
  }

  PRINT_HELP;
  return -1;
}
